/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.blog.nio.voipapplication.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;

import ch.maxant.blog.nio.voipapplication.common.Packet;
import ch.maxant.blog.nio.voipapplication.common.VoipProtocol;
import ch.maxant.blog.nio.voipapplication.server.model.VoipServerModel;

/**
 * a client which can talk and listen to other clients using the {@link VoipServerModel}.  
 */
public abstract class AbstractClient {

	private static final int INCOMING_CALL_DATA_BUFFER_SIZE_BYTES = 1000; //2500 bytes is ~ tenth of a second, when playing at 192 kbps
	
	protected String name;
	protected Socket socket;
	protected VoipClientProtocol protocol;
	protected boolean inCall = false;
	protected boolean quit = false; //set to true, when quitting, to exit the recv loop
	protected Thread recv;

	private void listenForCommands() {
		//hand off to other thread, because reading the response will block
		recv = new Thread(new Runnable(){

			@Override
			public void run() {
				long callDataReceived = 0L;
				long callStart = System.currentTimeMillis();
				final PipedInputStream inPipe = new PipedInputStream(10 * INCOMING_CALL_DATA_BUFFER_SIZE_BYTES); //must be bigger than our buffer, otherwise it stalls and we lose data!!
				PipedOutputStream outPipe = null;
				boolean startStreamingAudioWhenReady = true;
				while(!quit){
					try {
						if(outPipe == null) outPipe = new PipedOutputStream(inPipe);
						Packet response = protocol.readPacket(); //blocks until a command is sent
						if(response.getCommand() == VoipProtocol.LOGIN){
							System.out.println(name + ") unexpected login command at this time");
						}else if(response.getCommand() == VoipProtocol.TIME){
							System.out.println(name + ") " + response.getPayloadAsString());
						}else if(response.getCommand() == VoipProtocol.START_CALL){
							System.out.println(name + ") Call just started from " + response.getPayloadAsString());
							inCall = true;
							callDataReceived = 0L; //reset
							talk(); //send jibber jabber back to caller
						}else if(response.getCommand() == VoipProtocol.CALL_OK){
							System.out.println(name + ") Call connected");
							inCall = true;
							talk(); //send jibber jabber to callee
						}else if(response.getCommand() == VoipProtocol.CALL_NOK){
							System.out.println(name + ") Call failed: " + response.getPayloadAsString());
							inCall = false;
						}else if(response.getCommand() == VoipProtocol.CALL_DATA){
							callDataReceived += response.getPayload().length;
							//System.out.println(name + ") recvd call data: " + response.payload.length + ", total=" + callDataReceived);

							outPipe.write(response.getPayload()); //should pass it through to the input pipe which will pass it through to the audio input stream

							if(startStreamingAudioWhenReady && callDataReceived > INCOMING_CALL_DATA_BUFFER_SIZE_BYTES){
								System.out.println(name + ") sending buffered data for handling");
								startStreamingAudioWhenReady = false;
								handleCallData(inPipe);
								System.out.println(name + ") buffered data sent for handling");
							}
							
						}else if(response.getCommand() == VoipProtocol.END_CALL){
							//quite the call
							inCall = false;
							long time = System.currentTimeMillis() - callStart;
							System.out.println(name + ") Call ended.  Received " + (callDataReceived/1024.0) + " KBytes in " + (time/1000.0) + " seconds, averaging " + (1000*8.0*callDataReceived/1024.0/time) + " kbits/second");
						}else if(response.getCommand() == VoipProtocol.QUIT){
							System.out.println(name + ") Bye.");
							quit = true;
							socket.close();
						}else{
							System.out.println(name + ") Unknown command: (" + response.getCommand() + ")" + response.getPayloadAsString());
						}
					} catch (SocketException e) {
						//this sockets fucked.  prolly coz we're exiting - so let this thread run to its end.
						System.err.println("socket exception: " + e); //TODO but whys this happen in the listener?
						quit = true;
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} 
				}
			}
		}, name + "-recv");
		
		recv.start();
	}

	/** default does nothing */
	protected void handleCallData(InputStream in){}

	/**
	 * used at the end of a call to hang up and tell the other party you arent talking anymore.
	 * @throws IOException
	 */
	public void hangup() throws IOException {
		if(!inCall){
			System.out.println(name + ") hanging up, yet not in call...");
		}
		protocol.hangup();
		inCall = false;
	}

	/** default impl does nothing */
	protected void talk(){}

	protected void quit() throws InterruptedException, IOException {
		quit = true;  
		//now send in a command which replies, so that the recv listener can quit.
		protocol.quit();
	}

	/**
	 * creates a client and logs them straight into the server, ready for action
	 * @param endpoint the server
	 * @param friendlyName login name
	 * @throws IOException in case somit goes wrong
	 */
	public AbstractClient(InetSocketAddress endpoint, String friendlyName) throws IOException{
		
		this.name = friendlyName;
		
		//connect
		socket = new Socket();
		socket.connect(endpoint, 1000);
		protocol = new VoipClientProtocol(socket);
		
		//login
		String sessId = protocol.login(name);
		System.out.println(name + ") Session: " + sessId);
		
		listenForCommands();
		
		protocol.getServerTime();
	}
	
}
