/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.blog.nio.voipapplication.client;

import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

/**
 * creates a voip client which logs into the server and simply waits to be called.
 * if plays any stream it gets as an mp3.
 * http://www.javalobby.org/java/forums/t18465.html
 */
public class ListeningClient extends AbstractClient {

	/** main method for starting client, no args required. */
	public static void main(String[] args) throws UnknownHostException, IOException, InterruptedException {

		final InetSocketAddress endpoint = new InetSocketAddress(InetAddress.getLocalHost(), 30032);
//		final InetSocketAddress endpoint = new InetSocketAddress("192.168.1.103", 30032);
		System.out.println(new SimpleDateFormat("HH:mm:ss.SSS").format(new Date()) + " Starting client");

		new ListeningClient(endpoint, "listener");
	}

	/** plays incoming stream as an MP3 */
	@Override
	protected void handleCallData(final InputStream in) {
		//start a new thread, so that this blocking operation does not cause the caller to block.
		
		new Thread(new Runnable(){
			@Override
			public void run() {
				try{
					AudioInputStream audioIn = AudioSystem.getAudioInputStream(in);
					AudioFormat baseFormat = audioIn.getFormat();
					AudioFormat decodedFormat = new AudioFormat(
						AudioFormat.Encoding.PCM_SIGNED, // Encoding to use
						baseFormat.getSampleRate(),	  // sample rate (same as base format)
						16,				  // sample size in bits (thx to Javazoom)
						baseFormat.getChannels(),	  // # of Channels
						baseFormat.getChannels()*2,	  // Frame Size
						baseFormat.getSampleRate(),	  // Frame Rate
						false				  // Big Endian
					);
					AudioInputStream decodedIn = AudioSystem.getAudioInputStream(decodedFormat, audioIn);					
					DataLine.Info info = new DataLine.Info(SourceDataLine.class, decodedFormat);
					SourceDataLine line = (SourceDataLine) AudioSystem.getLine(info);
					line.open(decodedFormat);

					//Finally, once we have the line open, we have to implement a buffering strategy for tunneling data from our inputstream to our line-out. Note this is much like writing a buffering strategy to copy between two files - the technique is certainly not unique to audio transfer.

					if(line != null) {
						byte[] data = new byte[4096]; // 4k is a reasonable transfer size. TODO make smaller to improve latency?
						// Start
						line.start(); // Start the line.
						
						int nBytesRead;
						while ((nBytesRead = decodedIn.read(data, 0, data.length)) != -1) {	
							line.write(data, 0, nBytesRead);
						}
						// Stop
						line.drain();
						line.stop();
						line.close();
						decodedIn.close();
					}
				}catch(IOException e){
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (LineUnavailableException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (UnsupportedAudioFileException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}, name + "-audio-out").start();
	}

	/**
	 * @see AbstractClient#AbstractClient(InetSocketAddress, String)
	 */
	public ListeningClient(InetSocketAddress endpoint, String friendlyName) throws IOException{
		super(endpoint, friendlyName);
	}
	
}
