/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.blog.nio.voipapplication.client;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;

import ch.maxant.blog.nio.voipapplication.server.model.Participant;
/**
 * creates a voip client which calls another {@link Participant}.
 * to simulate the call, it streams an mp3 file to the server so others can listen.
 */
public class SendingClient extends AbstractClient {

	private static final long MS_TO_SEND_DATA = 20000L;
	private static final int NUM_BYTES_PER_PACKET = 100; // 100 = 800bits/packet = 12.2 milliseconds worth of data
	private static final int REQUIRED_DATA_RATE_BS = 192 * 1024; //64*1024 kbits per second = 65536 bits/s - use 192 here, coz my mp3 has that rate

	/** main method for starting client, no args required. */
	public static void main(String[] args) throws UnknownHostException, IOException, InterruptedException {

		final InetSocketAddress endpoint = new InetSocketAddress(InetAddress.getLocalHost(), 30032);
//		final InetSocketAddress endpoint = new InetSocketAddress("192.168.1.103", 30032);
		System.out.println(new SimpleDateFormat("HH:mm:ss.SSS").format(new Date()) + " Starting client");

		SendingClient c = new SendingClient(endpoint, "sender");

		c.call("listener");
	}

	private void call(final String recipient) throws IOException {
		System.out.println(name + ") initiating call to " + recipient);
		protocol.call(recipient);
		System.out.println(name + ") call request sent");
	}
	
	@Override
	protected void talk() {
		
		//start streaming data at correct rate for given time, then quit
		Thread send = new Thread(new Runnable(){
			@Override
			public void run() {
				try {
					
					File f = new File("d:/music/DuranDuranAllyouneedisnow01.mp3");
					if(!f.exists()) throw new IOException("unable to find file " + f.getAbsolutePath());
					InputStream input = new FileInputStream(f);
					input = new BufferedInputStream(input);
					
					System.out.println(name + " sending data for " + MS_TO_SEND_DATA + " ms...");
					//now we are set to go!
					long start = System.currentTimeMillis();
					long now = start;
					int numPacketsSent = 0;
					double avgWaitTimeTotal = 0.0;
					while(now < start + MS_TO_SEND_DATA){
						numPacketsSent++;

						byte[] data = new byte[NUM_BYTES_PER_PACKET];
						input.read(data);
//	for(int i = 0; i < data.length; i++){
//		data[i] = (byte) random.nextInt(Byte.MAX_VALUE);
//	}
						protocol.sendCallData(data);

						//pause, so that we do 64kbs, assuming that we are actually sending more than that!
						long timeRunSoFar = System.currentTimeMillis() - start; //eg 3ms
						long numBytesSoFar = numPacketsSent * NUM_BYTES_PER_PACKET; //eg 20,000
						long numBitsSoFar = numBytesSoFar * 8; //eg 160,000
						long timeExpectedAtRateForDataSoFar = 1000 * numBitsSoFar / REQUIRED_DATA_RATE_BS; //eg 160,000 / 64,000 = 2500ms
						long timeToWait = timeExpectedAtRateForDataSoFar - timeRunSoFar - 1; //eg 2500 - 3 -1 = 2496ms to wait (-1 to round down :-)
						if(timeToWait >= 0){
							try {
								avgWaitTimeTotal += timeToWait;
								Thread.sleep(timeToWait);
							} catch (InterruptedException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}else{
							//crap, we are falling behind!
							if(timeRunSoFar > 0){
								System.out.println(name + ") Warning: datarate is too low: " + (1000 * numBitsSoFar / timeRunSoFar / 1024) + " kbits per second (" + (REQUIRED_DATA_RATE_BS/1024) + " required)");
							}
						}

						//continue to next packet...
						now = System.currentTimeMillis();
					}
					
					//all done!
					double numKBytesSoFar = numPacketsSent * NUM_BYTES_PER_PACKET / 1024.0; //eg 20,000
					double avgWaitTime = avgWaitTimeTotal/numPacketsSent; //
					double msPerPacket = 1000.0 * 8 * NUM_BYTES_PER_PACKET / REQUIRED_DATA_RATE_BS; //eg 12ms
					System.out.println(name + ") finished sending " + numKBytesSoFar + " kilobytes in " + 
							numPacketsSent + " packets, " +
							"with an avg wait time of " + avgWaitTime + " ms (" + msPerPacket + " ms needed per packet, making load " + 
							(100.0 * (1.0-(avgWaitTime / msPerPacket))) + "%).  " +
							"averaged " + (1000*numKBytesSoFar*8)/(now-start) + " kbits/s.  " +
							"hanging up...");
					hangup();
					
					//in the real world, quitting would prolly happen when the
					//client is closed, rather than just at the end of the call.
					quit();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
			}
		}, name + "-send");

		send.start();
	}
	
	/**
	 * @see AbstractClient#AbstractClient(InetSocketAddress, String)
	 */
	public SendingClient(InetSocketAddress endpoint, String friendlyName) throws IOException{
		super(endpoint, friendlyName);
	}
	
}
