/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.blog.nio.framework.internal;

import java.io.IOException;
import java.security.Principal;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.annotation.security.RolesAllowed;

import ch.maxant.blog.nio.framework.Configuration;
import ch.maxant.blog.nio.framework.Handler;
import ch.maxant.blog.nio.framework.IHandler;
import ch.maxant.blog.nio.framework.ServerRequest;
import ch.maxant.blog.nio.framework.ServerResponse;
import ch.maxant.blog.nio.framework.UnknownHandlerException;

/**
 * the core of the server which handles the container in which handlers run.
 * uses reflection to cause {@link IHandler}s marked with {@link Handler#runAsync()} as true
 * to run in the execution service, at some time in the future (using a fixed size thread pool).
 */
public class Container {

	private Configuration config;
	private Principal user;
	private ExecutorService execService;
	private Map<String, CachedHandler> cachedHandlers;
	
	public Container(Configuration config) {
		this.config = config;
		this.execService = Executors.newFixedThreadPool(config.getMaxNumThreadsForAsync());
		cachedHandlers = Collections.synchronizedMap(new HashMap<String, CachedHandler>());
	}

	/**
	 * causes the handler to service the request in the thread pool, at some time in the future.
	 */
	@SuppressWarnings("rawtypes")
	private void runAsync(final IHandler ih, final ServerRequest request, final ServerResponse response) {
		this.execService.execute(new Runnable(){
			@Override
			public void run() {
				try {
					ih.service(request, response);
				} catch (IOException e) {
					//TODO how to handle errors properly?
					e.printStackTrace();
				}
			}
		});
	}
	
	public void handleRequest(String criterion, ServerRequest request, ServerResponse response) throws IOException, UnknownHandlerException {
		CachedHandler cachedHandler = cachedHandlers.get(criterion);
		if(cachedHandler == null){
			for(Class<? extends IHandler<?>> clazz : config.getHandlers()){
				Handler h = clazz.getAnnotation(Handler.class);
				if(h != null){
					if(h.selector().matches(criterion)){
						//TODO take from a pool?
						try {
							IHandler<?> ih = clazz.newInstance();
							cachedHandler = new CachedHandler(ih, h);
							cachedHandlers.put(criterion, cachedHandler);
							ih.setProtocol(config.getProtocol());
							break;
						} catch (InstantiationException e) {
							throw new IOException("unable to instantiate handler", e);
						} catch (IllegalAccessException e) {
							throw new IOException("unable to instantiate handler", e);
						}
					}
				}
			}
			if(cachedHandler == null){
				throw new UnknownHandlerException(criterion);
			}
		}
		//check security
		RolesAllowed ra = cachedHandler.ihandler.getClass().getAnnotation(RolesAllowed.class);
		if(ra != null && user != null){
			//TODO go check users roles against ra.value();
		}
		
		if(cachedHandler.annotation.runAsync()){
			runAsync(cachedHandler.ihandler, request, response);
		}else{
			cachedHandler.ihandler.service(request, response);
		}
	}

	private static class CachedHandler {
		IHandler<?> ihandler;
		Handler annotation;
		CachedHandler(IHandler<?> ihandler, Handler annotation){
			this.ihandler = ihandler;
			this.annotation = annotation;
			
		}
	}
	
}
