/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.blog.nio.framework.internal;

import java.io.IOException;
import java.lang.reflect.Method;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Iterator;
import java.util.Set;

import ch.maxant.blog.nio.framework.Configuration;
import ch.maxant.blog.nio.framework.Context;
import ch.maxant.blog.nio.framework.INonBlockingServer;
import ch.maxant.blog.nio.framework.IServerProtocol;
import ch.maxant.blog.nio.framework.ServerRequest;
import ch.maxant.blog.nio.framework.ServerResponse;

/**
 * with this one, I want to check true concurrency - or do i have to hand off to
 * thread to achieve it?
 */
public class NonBlockingServerImpl extends ServerStats implements INonBlockingServer {

	@SuppressWarnings("rawtypes")
	private final IServerProtocol protocol;
	
	private final Container container;
	
	private final Configuration config;
	
	public NonBlockingServerImpl(Configuration config){
		this.config = config;
		this.protocol = config.getProtocol();
		this.container = new Container(config);
		
		setContextInProtocol(config.getProtocol().getClass());
		
		System.out.println("Starting NIO server...");
		setupLogger();
	}
	
	private void setContextInProtocol(Class<? extends Object> clazz) {
		Method m;
		try {
			m = clazz.getDeclaredMethod("setContext", Context.class);
			m.setAccessible(true);
			m.invoke(config.getProtocol(), new ContextImpl(container));
		} catch (NoSuchMethodException e) {
			setContextInProtocol(clazz.getSuperclass());
		} catch (Exception e) {
			System.err.println("Unable to set context on protocol");
			e.printStackTrace();
		}
	}

	public void listen() throws IOException {

		Selector selector = Selector.open();

		ServerSocketChannel server = ServerSocketChannel.open();
		server.socket().bind(new InetSocketAddress(config.getPort()));
		server.configureBlocking(false);
		SelectionKey serverkey = server.register(selector, SelectionKey.OP_ACCEPT);

		ByteBuffer buffer = ByteBuffer.allocate(protocol.getBufferSize());

		boolean quit = false;
		long lastRecordedTime = System.currentTimeMillis();
		while (!quit) {

			lastRecordedTime = recordBusyTime(lastRecordedTime);
			
			selector.select(); // blocks until something arrives, of type OP_ACCEPT
			
			lastRecordedTime = recordIdleTime(lastRecordedTime);

			Set<SelectionKey> keys = selector.selectedKeys();

			for (Iterator<SelectionKey> i = keys.iterator(); i.hasNext();) {
				SelectionKey key = i.next();
				i.remove(); // extremely important to remove the key! otherwise
							// it uses up CPU cycles

				if (key == serverkey) {
					//its an op_accept key - a new request has arrived
					
					if (key.isAcceptable()) {
						SocketChannel client = server.accept();
						if (client != null) { 
							// can be null if theres no pending connection
							
							client.configureBlocking(false);
							
							// register for the read event
							client.register(selector, SelectionKey.OP_READ); 
							
							numConns++;
						}
					}
				} else {
					SocketChannel client = (SocketChannel) key.channel();
					if (!key.isReadable()) {
						continue;
					}
					try{
						boolean closeClient = false;
						try {
							int bytesread = client.read(buffer);
							if (bytesread == -1) {
								// when does this happen?
								closeClient = true;
							}
						} catch (IOException e) {
							closeClient = true;
						}
						if (closeClient) {
							protocol.closeClient(client, key);
							closeClient(); //removes from stats
						}else{
							buffer.flip();
							byte[] data = new byte[buffer.remaining()];
							System.arraycopy(buffer.array(), 0, data, 0, data.length);
							buffer.clear();

							numBytesProcessed += data.length;
							
							//buffer is now cleared so the protocol is free to handle the data on a new thread if it likes, and we are
							//free to handle the next incoming event
							incrementRequests();
							protocol.handleRequest(new ServerRequest(data, key), new ServerResponse(client));
						}
					}catch(Exception e){
						System.err.println("bit of a problem handling request: " + e);
						e.printStackTrace();
					}
				}
			}
		}
		System.out.println("done");
	}

}
